from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from cuppa.components.passthrough import PassthroughTransformer
from cuppa.logger import LoggerMixin

if TYPE_CHECKING:
    from cuppa.classifier.cuppa_classifier import CuppaClassifier
    from cuppa.compose.pipeline import Pipeline
    from cuppa.compose.column_transformer import ColumnTransformer


class BypassedClassifierBuilder(LoggerMixin):
    def __init__(
        self,
        cuppa_classifier: CuppaClassifier,
        verbose: bool = True
    ):
        self.cuppa_classifier = cuppa_classifier
        self.verbose = verbose

    _RECIPES = dict()

    _RECIPES["calibration"] = [
        ["meta_clfs", "dna_combined", "calibrator"],
        ["meta_clfs", "rna_combined", "calibrator"],
    ]

    _RECIPES["fusion_override"] = [
        ["meta_clfs", "dna_combined", "fusion_overrider"],
    ]

    _RECIPES["sex_filter"] = [
        ["meta_clfs", "dna_combined", "sex_filter"],
        ["meta_clfs", "rna_combined", "sex_filter"],
    ]

    _RECIPES["dna_only"] = [
        ["sub_clfs", "gen_pos"],
        ["sub_clfs", "snv96"],
        ["sub_clfs", "event"],
        # ["sub_clfs", "gene_exp"],
        # ["sub_clfs", "alt_sj"],
        ["meta_clfs", "rna_combined"],
        ["meta_clfs", "rna_combined"],
    ]

    @staticmethod
    def _get_parent_transformer(
        cuppa_classifier: CuppaClassifier,
        keys: list[str]
    ) -> Pipeline | ColumnTransformer:
        """
        Traverse the tree structure of a CuppaClassifier object to retrieve the estimator object that is one level up
        from the specified keys

        `_get_parent_estimator(cuppa_classifier, keys=["sub_clfs", "event", "feature_transform"])`

        is equivalent to:

        `cuppa_classifier["sub_clfs"]["event"]`
        """

        n_keys = len(keys)

        if n_keys == 1:
            return cuppa_classifier

        current_transformer = cuppa_classifier
        for key in keys[:-1]:
            current_transformer = current_transformer[key]

        return current_transformer

    @staticmethod
    def _bypass_transformer(
        composer: Pipeline | ColumnTransformer,
        transformer_name: str
    ) -> Pipeline | ColumnTransformer:
        """
        Within `Pipeline` or `ColumnTransformer` objects, replace a target transformer (by name) with a
        `PassthroughTransformer()`
        """

        transformer_names = [name for name, _ in composer.steps]
        if transformer_name not in transformer_names:
            raise LookupError("'%s' not found" % transformer_name)

        transformer_index = transformer_names.index(transformer_name)
        composer.steps[transformer_index] = (transformer_name, PassthroughTransformer())

        return composer

    def build(self, recipe: str, inplace: bool = False) -> CuppaClassifier:

        if inplace:
            cuppa_classifier = self.cuppa_classifier
        else:
            cuppa_classifier = copy.deepcopy(self.cuppa_classifier)

        keys_list = self._RECIPES[recipe]

        for keys in keys_list:
            #keys=keys_list[0]
            self._bypass_transformer(
                composer=self._get_parent_transformer(cuppa_classifier, keys),
                transformer_name=keys[-1]
            )

        return cuppa_classifier

    # @staticmethod
    # def get_estimator(cuppa_classifier, *keys: Iterable[str]):
    #     """
    #     Traverse the tree structure of a CuppaClassifier object to retrieve a specific estimator object
    #
    #     Calling `cuppa_classifier.get_estimator("sub_clfs", "gen_pos")` is equivalent to
    #     `cuppa_classifier["sub_clfs"]["gen_pos"]`
    #
    #     Parameters
    #     ----------
    #     keys : Iterable[str]
    #         names of estimators that can be passed to the `__get_item__()` method from
    #         `cuppa.compose.column_transformer.ColumnTransformer` or `cuppa.compose.column_transformer.Pipeline`
    #
    #     Returns
    #     -------
    #     sklearn.base.BaseEstimator
    #     """
    #
    #     current_estimator = cuppa_classifier
    #
    #     if len(keys) == 1:
    #         return current_estimator[keys[0]]
    #
    #     for key in keys:
    #         current_estimator = current_estimator[key]
    #
    #     return current_estimator


    # def _check_bypass_steps(self) -> None:
    #     valid_bypass_steps = ["calibrator", "fusion_overrider", "sex_filter"]
    #
    #     if self.bypass_steps == "all":
    #         self.bypass_steps = valid_bypass_steps
    #
    #     if ~pd.Series(self.bypass_steps).isin(valid_bypass_steps).all():
    #         self.logger.error("Valid `bypass_steps` are: " + ", ".join(valid_bypass_steps) + ", or all")
    #         raise ValueError
    #
    # def _bypass_step(self, pipeline, step: str) -> None:
    #     step_names = [step_i for step_i, transformer in pipeline.steps]
    #     if step not in step_names:
    #         self.logger.error("'%s' not found" % step)
    #         raise LookupError
    #
    #     step_index = step_names.index(step)
    #     pipeline.steps[step_index] = (step, PassthroughTransformer())
    #
    # def bypass_prob_calibration(self, cuppa_classifier: CuppaClassifier) -> None:
    #     if self.verbose:
    #         self.logger.info("Removing probability calibrators")
    #
    #     self._bypass_step(cuppa_classifier.dna_combined_clf, "calibrator")
    #     self._bypass_step(cuppa_classifier.rna_combined_clf, "calibrator")
    #
    # def bypass_fusion_overrider(self, cuppa_classifier: CuppaClassifier) -> None:
    #     if self.verbose:
    #         self.logger.info("Removing fusion overrider")
    #
    #     self._bypass_step(cuppa_classifier.dna_combined_clf, "fusion_overrider")
    #
    # def bypass_sex_filters(self, cuppa_classifier: CuppaClassifier) -> None:
    #     if self.verbose:
    #         self.logger.info("Removing sex filter")
    #
    #     self._bypass_step(cuppa_classifier.dna_combined_clf, "sex_filter")
    #     self._bypass_step(cuppa_classifier.rna_combined_clf, "sex_filter")
    #
    # def build(self):
    #
    #     if self.inplace:
    #         if self.verbose:
    #             self.logger.info("Modifying CuppaClassifier in place")
    #         cuppa_classifier = self.cuppa_classifier
    #     else:
    #         cuppa_classifier = copy.deepcopy(self.cuppa_classifier)
    #
    #     if "calibrator" in self.bypass_steps:
    #         self.bypass_prob_calibration(cuppa_classifier)
    #
    #     if "fusion_overrider" in self.bypass_steps:
    #         self.bypass_fusion_overrider(cuppa_classifier)
    #
    #     if "sex_filter" in self.bypass_steps:
    #         self.bypass_sex_filters(cuppa_classifier)
    #
    #     return cuppa_classifier
